#include "./puzzram.h"

/***********************************************************/
/* MENU ROUTINES                                           */
/***********************************************************/
/****************************************/
/* Main window display.                 */
/****************************************/
void MAIN_WINDOW(void) {
   glDrawBuffer(GL_BACK);
   glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
   glColor3f(0.6, 0.6, 0.6);
   glBegin(GL_QUADS);
     glVertex2i(  0,   0);
     glVertex2i(640,   0);
     glVertex2i(640, 480);
     glVertex2i(  0, 480);
   glEnd();
   // Draw etched lines.
   glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
   glColor3f(0.8, 0.8, 0.8);
   glBegin(GL_LINES);
     glVertex2i(  0, 440);
     glVertex2i(649, 440);
     glVertex2i(  0,   5);
     glVertex2i(649,   5);
     glVertex2i(  0, 200);
     glVertex2i(110, 200);
     glVertex2i(110, 440);
     glVertex2i(110,   5);
   glEnd();
   glColor3f(0.0, 0.0, 0.0);
   glBegin(GL_LINES);
     glVertex2i(  0, 441);
     glVertex2i(649, 441);
     glVertex2i(  0,   6);
     glVertex2i(649,   6);
     glVertex2i(  0, 201);
     glVertex2i(110, 201);
     glVertex2i(109, 440);
     glVertex2i(109,   6);
   glEnd();
   glutSwapBuffers();
   // Draw windows.
   PLAY_WINDOW.DRAW();
   PLAY_WINDOW.CLEAR();
   TEXT_COPYRIGHT();
   MOVE_WINDOW.DRAW();
   MOVE_WINDOW.CLEAR();
   TEXT_MOVE_COUNT();
   LEVL_WINDOW.DRAW();
   LEVL_WINDOW.CLEAR();
   TEXT_DIFFICULTY();
   HELP_WINDOW.DRAW();
   HELP_WINDOW.CLEAR();
   // Draw buttons.
   TEXT_8_X_13( 20, 420, "Game Menu");
   NEWG_BUTTON.DRAW();
   LVLP_BUTTON.DRAW();
   LVLN_BUTTON.DRAW();
   PLAY_BUTTON.DRAW();
   LOAD_BUTTON.DRAW();
   SAVE_BUTTON.DRAW();
   EXIT_BUTTON.DRAW_HILITE(1.0, 0.5, 0.5);
   TEXT_8_X_13( 20, 180, "Help Menu");
   LABL_BUTTON.DRAW();
   VIEW_BUTTON.DRAW();
}

/****************************************/
/* Draws copyright data on screen.      */
/****************************************/
void TEXT_COPYRIGHT(void) {
char text[80];
   strcpy(text, "PUZZ-R.A.M. (");
   strcat(text, version);
   strcat(text, ")");
   TEXT_8_X_13_WHITE(250, 240, text);
   TEXT_8_X_13_WHITE(250, 220, "Copyright (C), Ronald Marsh, 2001.");
   TEXT_8_X_13_WHITE(250, 200, "All rights reserved.");
}

/****************************************/
/* Draws # of puzzle moves on screen.   */
/****************************************/
void TEXT_MOVE_COUNT(void) {
char text[20];
char move[10];
   sprintf(move, "%d", movesMade);
   strcpy(text, "Moves:");
   strcat(text, move);
   MOVE_WINDOW.CLEAR();
   TEXT_8_X_13_WHITE( 15, 455, text);
   glutSwapBuffers();
}

/****************************************/
/* Draws difficulty level on screen.    */
/****************************************/
void TEXT_DIFFICULTY(void) {
char text[20];
char level[10];
   sprintf(level, "%d", difficulty);
   strcpy(text, "Level:");
   strcat(text, level);
   LEVL_WINDOW.CLEAR();
   TEXT_8_X_13_WHITE(541, 455, text);
   glutSwapBuffers();
}

/***********************************************************/
/* DISK SCAN ROUTINES                                      */
/***********************************************************/
/****************************************/
/* PGM file scan sub-window display.    */
/****************************************/
void SCAN_WINDOW(void) {
   glDrawBuffer(GL_BACK);
   glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
   glColor3f(0.8, 0.8, 0.8);
   glBegin(GL_QUADS);
     // Window top border.
     glVertex2i(225, 420);
     glVertex2i(525, 420);
     glVertex2i(523, 418);
     glVertex2i(227, 418);
     // Window left border.
     glVertex2i(225, 420);
     glVertex2i(227, 418);
     glVertex2i(227,  24);
     glVertex2i(225,  20);
   glEnd();
   glColor3f(0.4, 0.4, 0.4);
   glBegin(GL_QUADS);
     // Window bottom border.
     glVertex2i(227,  22);
     glVertex2i(225,  20);
     glVertex2i(525,  20);
     glVertex2i(523,  22);
     // Window right border.
     glVertex2i(523, 418);
     glVertex2i(525, 420);
     glVertex2i(525,  20);
     glVertex2i(523,  22);
   glEnd();
   glColor3f(0.6, 0.6, 0.6);
   glBegin(GL_QUADS);
     glVertex2i(227, 418);
     glVertex2i(523, 418);
     glVertex2i(523,  22);
     glVertex2i(227,  22);
   glEnd();
   glutSwapBuffers();
   // Draw window.
   FILE_WINDOW.DRAW();
   FILE_WINDOW.CLEAR();
   // Draw buttons.
   UP___BUTTON.DRAW();
   DOWN_BUTTON.DRAW();
   CNCL_BUTTON.DRAW();
}

/****************************************/
/* PGM file scan.                     	*/
/****************************************/
int SCAN_FOR_IMAGES(char Extension[4]) {
int  i;
char sysCmd[80];
FILE *disk;
   // Scan disk for files.
   strcpy(sysCmd, "ls *");
   strcat(sysCmd, Extension);
   strcat(sysCmd, " > puzzleList");
   system(sysCmd);
   if ((disk = fopen("puzzleList", "r")) != NULL) {
      for (i = 0; ; i++) {
          fscanf(disk, "%s", PuzzleNames[i]);
          if (feof(disk)) break;
      }
   }
   fclose(disk);
   system("rm puzzleList");
   if (i == 0) return 0;
          else return --i;
}

/****************************************/
/* PGM file scan display.               */
/****************************************/
void SCAN_FILE_DISPLAY(int start, int puzzles) {
int i, j = start;
   FILE_WINDOW.CLEAR();
   for (i = 0; i < 350; i+=20) {
       TEXT_8_X_13_WHITE( 245, (385-i), PuzzleNames[j]);
       if (j == puzzles) break;
       j++;
   }
   glutSwapBuffers();
}


/***********************************************************/
/* NEW PUZZLE ROUTINES                                     */
/***********************************************************/
/****************************************/
/* PGM file load (new puzzle image).    */  
/****************************************/
int NEW_LOAD(char FileName[80]) { 
int   x, y;
int   xmax, ymax, pmax;
char  ch;
char  type[3];
char  command[256];
FILE *disk;
   if ((disk = fopen(FileName, "rb")) == NULL) {
      TEXT_8_X_13_WHITE(250, 220, "ERROR:");
      TEXT_8_X_13_WHITE(250, 200, "- The file can not be opened.");
      glutSwapBuffers();
      return 1;
   }
   // Read in file header.
   fscanf(disk, "%s", &type);
   fscanf(disk, "%d", &xmax);
   fscanf(disk, "%d", &ymax);
   fscanf(disk, "%d", &pmax);
   fread(&ch, 1, 1, disk);
   // Check file size.
   if (xmax != puzzleW || ymax != puzzleH) {
      TEXT_8_X_13_WHITE(250, 220, "ERROR:");
      TEXT_8_X_13_WHITE(250, 200, "- The file is not sized correctly.");
      TEXT_8_X_13_WHITE(250, 180, "  It must be 520 (w) by 425 (h).");
      fclose(disk);
      glutSwapBuffers();
      return 1;
   }

   // Load B/W image data.
   if (strcmp(type, "P5") == 0) {
      for (y = 0; y < ymax; y++) {
          for (x = 0; x < xmax; x++) {
              PuzzleData[y][x].r = getc(disk);
              PuzzleData[y][x].g = PuzzleData[y][x].r;
              PuzzleData[y][x].b = PuzzleData[y][x].r;
          }
      }
      fclose(disk);
   // Load color image.
   } else if (strcmp(type, "P6") == 0) {  
      for (y = 0; y < ymax; y++) {
          for (x = 0; x < xmax; x++) {
              PuzzleData[y][x].r = getc(disk);
              PuzzleData[y][x].g = getc(disk);
              PuzzleData[y][x].b = getc(disk);
          }
      }
      fclose(disk);
   }

   return 0;
}

/****************************************/
/* Draw puzzle image.                   */
/****************************************/
void NEW_DRAW(void) {
GLint xp, yp;
   glDrawBuffer(GL_BACK);
   glPointSize(2);
   glBegin(GL_POINTS);
   yp = PLAY_WINDOW.ys + 1;
   for (int y = puzzleH-1; y >= 0; y--) {
       xp = PLAY_WINDOW.xs + 1;
       for (int x = 0; x < puzzleW; x++) { 
           glColor3ub(PuzzleData[y][x].r, PuzzleData[y][x].g, PuzzleData[y][x].b);
           glVertex2i(xp, yp);
           xp++;
       }
       yp++;
   }
   glEnd();
   glPointSize(1);
   // Remove edges.
   glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
   glColor3f(0.0, 0.0, 0.0);
   glBegin(GL_QUADS);
     glVertex2i(PLAY_WINDOW.xs, PLAY_WINDOW.ys);
     glVertex2i(PLAY_WINDOW.xe, PLAY_WINDOW.ys);
     glVertex2i(PLAY_WINDOW.xe, PLAY_WINDOW.ye);
     glVertex2i(PLAY_WINDOW.xs, PLAY_WINDOW.ye);
   glEnd();
   glutSwapBuffers();
}

/****************************************/
/* Cut puzzle into pieces.              */
/****************************************/
void NEW_CUTUP_IMAGE(void) {
int   x, y;
int   piece = 0;
GLint xs, xe, ys, ye;
   glDrawBuffer(GL_BACK);
   glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
   for (y = 0; y < 5; y++) {
       ys = (PLAY_WINDOW.ys + 1 + (y * 85));
       ye = (ys + 85);
       for (x = 0; x < 5; x++) {
           xs = (PLAY_WINDOW.xs + 1 + (x * 104));
           xe = (xs  + 104);
           glColor3f(0.8, 0.8, 0.8);
           glBegin(GL_QUADS);
             // Piece top border.
             glVertex2i(xs,   ye);
             glVertex2i(xs+2, ye-2);
             glVertex2i(xe-2, ye-2);
             glVertex2i(xe,   ye);
             // Piece left border.
             glVertex2i(xs,   ye);
             glVertex2i(xs,   ys);
             glVertex2i(xs+2, ys+2);
             glVertex2i(xs+2, ye-2);
           glEnd();
           glColor3f(0.4, 0.4, 0.4);
           glBegin(GL_QUADS);
             // Piece bottom border.
             glVertex2i(xs,   ys);
             glVertex2i(xe,   ys);
             glVertex2i(xe-2, ys+2);
             glVertex2i(xs+2, ys+2);
             // Piece right border.
             glVertex2i(xe,   ye);
             glVertex2i(xe-2, ye-2);
             glVertex2i(xe-2, ys+2);
             glVertex2i(xe,   ys);
           glEnd(); 
           // Record puzzle piece locations.
           PuzzleBoard[y][x] = piece++;
       }
   }
   // Remove bottom-left corner piece.
   glColor3f(0.0, 0.0, 0.0);
   glBegin(GL_QUADS);
     glVertex2i(PLAY_WINDOW.xe-105, PLAY_WINDOW.ys);
     glVertex2i(PLAY_WINDOW.xe, PLAY_WINDOW.ys);
     glVertex2i(PLAY_WINDOW.xe, PLAY_WINDOW.ys+86);
     glVertex2i(PLAY_WINDOW.xe-105, PLAY_WINDOW.ys+86);
   glEnd();
   PuzzleBoard[4][4] = -1;
   glutSwapBuffers();
}

/****************************************/
/* Randomize pieces.                    */
/****************************************/
void NEW_RANDOMIZE_IMAGE(void) {
int   i;
int   x1, x2, y1, y2;
int   temp;
GLint xs1, ys1, xs2, ys2;
RGB  *pieceA = new RGB[85*104];
RGB  *pieceB = new RGB[85*104];
   srand((int)(time(NULL)%42));
   glDrawBuffer(GL_FRONT_AND_BACK);
   glPixelTransferi(GL_MAP_COLOR, false);
   for (i = 0; i < (difficulty*4); i++) {
       x1  = (int)(5.0 * rand() / (RAND_MAX + 1.0));
       y1  = (int)(5.0 * rand() / (RAND_MAX + 1.0));
       xs1 = (PLAY_WINDOW.xs + 1 + (x1 * 104));
       ys1 = (PLAY_WINDOW.ys + 1 + (y1 * 85));
       glReadPixels(xs1, ys1, 104, 85, GL_RGB, GL_UNSIGNED_BYTE, pieceA);
       x2  = (int)(5.0 * rand() / (RAND_MAX + 1.0));
       y2  = (int)(5.0 * rand() / (RAND_MAX + 1.0));
       xs2 = (PLAY_WINDOW.xs + 1 + (x2 * 104));
       ys2 = (PLAY_WINDOW.ys + 1 + (y2 * 85));
       glReadPixels(xs2, ys2, 104, 85, GL_RGB, GL_UNSIGNED_BYTE, pieceB);
       glRasterPos2i(xs2, ys2);
       glDrawPixels(104, 85, GL_RGB, GL_UNSIGNED_BYTE, pieceA);
       glRasterPos2i(xs1, ys1);
       glDrawPixels(104, 85, GL_RGB, GL_UNSIGNED_BYTE, pieceB);
       // Record new piece locations.
       y1 = (4 - y1);
       y2 = (4 - y2);
       temp                = PuzzleBoard[y1][x1];
       PuzzleBoard[y1][x1] = PuzzleBoard[y2][x2];
       PuzzleBoard[y2][x2] = temp;
   }
   delete[] pieceA;
   delete[] pieceB;
}


/***********************************************************/
/* PUZZLE IO ROUTINES                                      */
/***********************************************************/
/****************************************/
/* Save puzzle to disk.                 */
/****************************************/
void PUZZLE_SAVE(char FileName[80]) {
int   x, y;
char  FName[80];
GLint xs = PLAY_WINDOW.xs+1;
GLint ys = PLAY_WINDOW.ys+1;
GLint ye = PLAY_WINDOW.ye-1;
RGB  *slice = new RGB[puzzleW];
FILE *disk;
   // Save filename.
   strcpy(FName, FileName);
   strcat(FName, ".puz");
   disk = fopen(".recoverData", "w");
   fprintf(disk, "%s\n", FName);
   fclose(disk);
   // Write undo data.
   disk = fopen(FName, "wb");
   fprintf(disk, "%d\n", movesMade);
   fprintf(disk, "%d\n", movesToSolve);
   for (y = 0; y < 6; y++) {
       for (x = 0; x < movesToSolve; x++) {
           fprintf(disk, "%d\n", PuzzleMoves[y][x]);
       }
   }
   // Write game board data.
   fprintf(disk, "%d\n", difficulty);
   for (y = 0; y < 5; y++) {
       for (x = 0; x < 5; x++) {
           fprintf(disk, "%d\n", PuzzleBoard[y][x]);
       }
   }
   // Write game image.
   glPixelTransferi(GL_MAP_COLOR, false);
   for (y = ye; y >= ys; y--) {
       glReadPixels(xs, y, puzzleW, 1, GL_RGB, GL_UNSIGNED_BYTE, slice);
       for (x = 0; x < puzzleW; x++) {
           fprintf(disk, "%c%c%c", slice[x].r, slice[x].g, slice[x].b);
       }
   }
   delete[] slice;
   // Write game help image.
   for (y = 0; y < 74; y++) {
       for (x = 0; x < 90; x++) {
           fprintf(disk, "%c%c%c", PuzzleView[y][x].r, 
                                   PuzzleView[y][x].g, 
                                   PuzzleView[y][x].b);
       }
   }
   fclose(disk);
}

/****************************************/
/* Load puzzle from disk.               */
/****************************************/
void PUZZLE_LOAD(void) {
int   x, y;
char  FName[80];
GLint xs = PLAY_WINDOW.xs+1;
GLint ys = PLAY_WINDOW.ys+1;
GLint ye = PLAY_WINDOW.ye-1;
RGB  *slice = new RGB[puzzleW];
FILE *disk;
   // Recover saved filename.
   if ((disk = fopen(".recoverData", "r")) == NULL) return;
   fscanf(disk, "%s", FName);
   fclose(disk);
   // Read undo data.
   if ((disk = fopen(FName, "rb")) == NULL) return;
   fscanf(disk, "%d", &movesMade);
   fscanf(disk, "%d", &movesToSolve);
   for (y = 0; y < 6; y++) {
       for (x = 0; x < movesToSolve; x++) {
           fscanf(disk, "%d", &PuzzleMoves[y][x]);
       }
   }
   // Read game board data.
   fscanf(disk, "%d", &difficulty);
   for (y = 0; y < 5; y++) {
       for (x = 0; x < 5; x++) {
           fscanf(disk, "%d\n", &PuzzleBoard[y][x]);
       }
   }
   // Read game image.
   glDrawBuffer(GL_BACK);
   glPixelTransferi(GL_MAP_COLOR, false);
   for (y = ye; y >= ys; y--) {
       for (x = 0; x < puzzleW; x++) {
           fscanf(disk, "%c%c%c", &slice[x].r, &slice[x].g, &slice[x].b);
       }
       glRasterPos2i(xs, y);
       glDrawPixels(puzzleW, 1, GL_RGB, GL_UNSIGNED_BYTE, slice);
   }
   glutSwapBuffers();
   delete[] slice;
   // Read game help image.
   for (y = 0; y < 74; y++) {
       for (x = 0; x < 90; x++) {
           fscanf(disk, "%c%c%c", &PuzzleView[y][x].r, 
                                  &PuzzleView[y][x].g, 
                                  &PuzzleView[y][x].b);
       }
   }
   fclose(disk);
}


/***********************************************************/
/* RECOVERY ROUTINES                                       */
/***********************************************************/
/****************************************/
/* Record image behind sub-window.      */
/****************************************/
void RECOVER_RECORD(void) {
   glPixelTransferi(GL_MAP_COLOR, false);
   glReadPixels(225, 20, 300, 400, GL_RGB, GL_UNSIGNED_BYTE, RecoverData);
}

/****************************************/
/* Repaint image behind sub-window.     */
/****************************************/
void RECOVER_DRAW() {
GLint xp, yp;
   glDrawBuffer(GL_BACK);
   glPointSize(2);
   glBegin(GL_POINTS);
   yp = 20;
   for (int y = 0; y < 400; y++) {
       xp = 225;
       for (int x = 0; x < 300; x++) { 
           glColor3ub(RecoverData[y][x].r, RecoverData[y][x].g, RecoverData[y][x].b);
           glVertex2i(xp, yp);
           xp++;
       }
       yp++;
   }
   glEnd();
   glPointSize(1);
   glutSwapBuffers();
}


/***********************************************************/
/* PLAY PUZZLE ROUTINES                                    */
/***********************************************************/
/****************************************/
/* A winner.                            */
/****************************************/
void PLAY_WINNER(void) {
   if (winFlag == 0) return;
   HELP_WINDOW.CLEAR(); 
   usleep(500000);
   TEXT_8_X_13_WHITE( 18, 70, "You ");
   usleep(500000);
   TEXT_8_X_13_WHITE( 50, 70, "are ");
   usleep(500000);
   TEXT_8_X_13_WHITE( 82, 70, "a");
   usleep(500000);
   TEXT_8_X_13_WHITE( 26, 40, "WINNER!");
   usleep(500000);
   usleep(500000);
}

/****************************************/
/* Is it a winner?                      */
/****************************************/
void PLAY_ISWINNER(int x, int y) {
int winner;
   winner = 1;
   for (y = 0; y < 5; y++) if (PuzzleBoard[y][0] != (0 + (y * 5))) winner = 0;
   for (y = 0; y < 5; y++) if (PuzzleBoard[y][1] != (1 + (y * 5))) winner = 0;
   for (y = 0; y < 5; y++) if (PuzzleBoard[y][2] != (2 + (y * 5))) winner = 0;
   for (y = 0; y < 5; y++) if (PuzzleBoard[y][3] != (3 + (y * 5))) winner = 0;
   for (y = 0; y < 4; y++) if (PuzzleBoard[y][4] != (4 + (y * 5))) winner = 0;
   if (winner == 1) {
      winFlag = 1;
      PLAY_WINNER();
   }
}

/****************************************/
/* Play puzzle.                         */
/****************************************/
void PLAY(int x, int y) {
int   i, j;
int   x1, x2, y1, y2;
int   moveX = 0;
int   moveY = 0;
int   temp;
GLint xp = PLAY_WINDOW.xs + 1;
GLint yp = PLAY_WINDOW.ys + 1;
GLint xP, yP, xB, yB;
RGB  *piece = new RGB[85*104];
   glDrawBuffer(GL_FRONT_AND_BACK);
   // Determine piece picked.
   x  = ((x - 116) / 104);
   y  = ((y -  10) / 85);
   x1 = x;
   y1 = (4 - y);
   // Determine if piece can move right.
   j = 0;
   for (i = x; i < 5; i++) {
       if (PuzzleBoard[y1][i] == -1) moveX = j;
       j++;
   }
   // Move piece(s).
   while (moveX != 0) {
       yP = (yp + (y * 85));
       xP = (xp + ((x + moveX) * 104));
       yB = yP;
       xB = (xP - 104);
       moveX--;
       glPixelTransferi(GL_MAP_COLOR, false);
       glReadPixels(xB, yB, 104, 85, GL_RGB, GL_UNSIGNED_BYTE, piece);
       glRasterPos2i(xP, yP);
       glDrawPixels(104, 85, GL_RGB, GL_UNSIGNED_BYTE, piece);
       REDRAW_BLANK_PIECE(xB, yB);
       // Record new piece locations.
       x1 = (x + moveX + 1);
       x2 = (x + moveX);
       y1 = (4 - y);
       y2 = (4 - y);
       temp                = PuzzleBoard[y1][x1];
       PuzzleBoard[y1][x1] = PuzzleBoard[y2][x2];
       PuzzleBoard[y2][x2] = temp;
       // Record moves.
       PuzzleMoves[0][movesToSolve] = y;
       PuzzleMoves[1][movesToSolve] = x1;
       PuzzleMoves[2][movesToSolve] = PuzzleBoard[y1][x1];
       PuzzleMoves[3][movesToSolve] = y;
       PuzzleMoves[4][movesToSolve] = x2;
       PuzzleMoves[5][movesToSolve] = PuzzleBoard[y2][x2];
       // Update moves counters.
       movesToSolve++;
       movesMade++;
       TEXT_MOVE_COUNT();
       // Do we have a winner?
       PLAY_ISWINNER(x, y);
   }

   // Determine if piece can move left.
   j = 0;
   for (i = x; i >= 0; i--) {
       if (PuzzleBoard[y1][i] == -1) moveX = j;
       j--;
   }
   // Move piece(s).
   while (moveX != 0) {
       yP = (yp + (y * 85));
       xP = (xp + ((x + moveX) * 104));
       yB = yP;
       xB = (xP + 104);
       moveX++;
       glPixelTransferi(GL_MAP_COLOR, false);
       glReadPixels(xB, yB, 104, 85, GL_RGB, GL_UNSIGNED_BYTE, piece);
       glRasterPos2i(xP, yP);
       glDrawPixels(104, 85, GL_RGB, GL_UNSIGNED_BYTE, piece);
       REDRAW_BLANK_PIECE(xB, yB);
       // Record new piece locations.
       x1 = (x + moveX - 1);
       x2 = (x + moveX);
       y1 = (4 - y);
       y2 = (4 - y);
       temp                = PuzzleBoard[y1][x1];
       PuzzleBoard[y1][x1] = PuzzleBoard[y2][x2];
       PuzzleBoard[y2][x2] = temp;
       // Record moves.
       PuzzleMoves[0][movesToSolve] = y;
       PuzzleMoves[1][movesToSolve] = x1;
       PuzzleMoves[2][movesToSolve] = PuzzleBoard[y1][x1];
       PuzzleMoves[3][movesToSolve] = y;
       PuzzleMoves[4][movesToSolve] = x2;
       PuzzleMoves[5][movesToSolve] = PuzzleBoard[y2][x2];
       // Update moves counters.
       movesToSolve++;
       movesMade++;
       TEXT_MOVE_COUNT();
       // Do we have a winner?
       PLAY_ISWINNER(x, y);
   }

   // Determine if piece can move down.
   j = 0;
   for (i = (4-y); i < 5; i++) {
       if (PuzzleBoard[i][x1] == -1) moveY = j;
       j--;
   }
   // Move piece(s).
   while (moveY != 0) {
       yP = (yp + ((y + moveY) * 85));
       xP = (xp + (x * 104));
       yB = (yP + 85); 
       xB = xP;
       moveY++;
       glPixelTransferi(GL_MAP_COLOR, false);
       glReadPixels(xB, yB, 104, 85, GL_RGB, GL_UNSIGNED_BYTE, piece);
       glRasterPos2i(xP, yP);
       glDrawPixels(104, 85, GL_RGB, GL_UNSIGNED_BYTE, piece);
       REDRAW_BLANK_PIECE(xB, yB);
       // Record new piece locations.
       x1 = x;
       x2 = x;
       y1 = (4 - y) - (moveY - 1);
       y2 = (4 - y) - moveY;
       temp                = PuzzleBoard[y1][x1];
       PuzzleBoard[y1][x1] = PuzzleBoard[y2][x2];
       PuzzleBoard[y2][x2] = temp;
       // Record moves.
       PuzzleMoves[0][movesToSolve] = 4 - y1;
       PuzzleMoves[1][movesToSolve] = x;
       PuzzleMoves[2][movesToSolve] = PuzzleBoard[y1][x1];
       PuzzleMoves[3][movesToSolve] = 4 - y2;
       PuzzleMoves[4][movesToSolve] = x;
       PuzzleMoves[5][movesToSolve] = PuzzleBoard[y2][x2];
       // Update moves counters.
       movesToSolve++;
       movesMade++;
       TEXT_MOVE_COUNT();
       // Do we have a winner?
       PLAY_ISWINNER(x, y);
   }

   // Determine if piece can move up.
   j = 0;
   for (i = (4-y); i >= 0; i--) {
       if (PuzzleBoard[i][x1] == -1) moveY = j;
       j++;
   }
   // Move piece(s).
   while (moveY != 0) {
       yP = (yp + ((y + moveY) * 85));
       xP = (xp + (x * 104));
       yB = (yP - 85); 
       xB = xP;
       moveY--;
       glPixelTransferi(GL_MAP_COLOR, false);
       glReadPixels(xB, yB, 104, 85, GL_RGB, GL_UNSIGNED_BYTE, piece);
       glRasterPos2i(xP, yP);
       glDrawPixels(104, 85, GL_RGB, GL_UNSIGNED_BYTE, piece);
       REDRAW_BLANK_PIECE(xB, yB);
       // Record new piece locations.
       x1 = x;
       x2 = x;
       y1 = (4 - y) - (moveY + 1);
       y2 = (4 - y) - moveY;
       temp                = PuzzleBoard[y1][x1];
       PuzzleBoard[y1][x1] = PuzzleBoard[y2][x2];
       PuzzleBoard[y2][x2] = temp;
       // Record moves.
       PuzzleMoves[0][movesToSolve] = 4 - y1;
       PuzzleMoves[1][movesToSolve] = x;
       PuzzleMoves[2][movesToSolve] = PuzzleBoard[y1][x1];
       PuzzleMoves[3][movesToSolve] = 4 - y2;
       PuzzleMoves[4][movesToSolve] = x;
       PuzzleMoves[5][movesToSolve] = PuzzleBoard[y2][x2];
       // Update moves counters.
       movesToSolve++;
       movesMade++;
       TEXT_MOVE_COUNT();
       // Do we have a winner?
       PLAY_ISWINNER(x, y);
   }
   delete[] piece;
}

/****************************************/
/* Redraw blank piece.                  */
/****************************************/
void REDRAW_BLANK_PIECE(GLint x, GLint y) {
   glDrawBuffer(GL_BACK);
   glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
   glColor3f(0.0, 0.0, 0.0);
   glBegin(GL_QUADS);
     glVertex2i(x,     y);
     glVertex2i(x+104, y);
     glVertex2i(x+104, y+85);
     glVertex2i(x,     y+85);
   glEnd();
   glutSwapBuffers();
}


/***********************************************************/
/* HELP PUZZLE ROUTINES                                    */
/***********************************************************/
/****************************************/
/* Record puzzle help view              */
/****************************************/
void NEW_HELPVIEW_RECORD(void) {
int x = PLAY_WINDOW.xs + 1;
int y = PLAY_WINDOW.ys + 1;
RGB PuzzleTemp[puzzleH*puzzleW];
   glPixelTransferi(GL_MAP_COLOR, false);
   glReadPixels(  x,  y,   puzzleW,   puzzleH, GL_RGB, GL_UNSIGNED_BYTE, PuzzleTemp);
   gluScaleImage(GL_RGB,   puzzleW,   puzzleH, GL_UNSIGNED_BYTE, PuzzleTemp,
                         helpViewW, helpViewH, GL_UNSIGNED_BYTE, PuzzleView);
}

/****************************************/
/* Draw sub-puzzle.                     */
/****************************************/
void HELP_VIEW(void) {
int x = HELP_WINDOW.xs + 1;
int y = HELP_WINDOW.ys + 1;
   glDrawBuffer(GL_FRONT_AND_BACK);
   glPixelTransferi(GL_MAP_COLOR, false);
   glRasterPos2i(x, y);
   glDrawPixels(helpViewW, helpViewH, GL_RGB, GL_UNSIGNED_BYTE, PuzzleView);
}

/****************************************/
/* Number pieces.                       */
/****************************************/
void HELP_NUMBER(void) {
int   i, j;
char  number[4];
GLint xp, yp;
   glDrawBuffer(GL_BACK);
   for (j = 0; j < 5; j++) { 
       yp = (431 - (j * 85));
       for (i = 0; i < 5; i++) { 
           xp = (119 + (i * 104));
           // Space.
           if (PuzzleBoard[j][i] != -1) {
              glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
              glColor3f(0.0, 0.0, 0.0);
              glBegin(GL_QUADS);
                glVertex2i(xp,    yp);
                glVertex2i(xp+20, yp);
                glVertex2i(xp+20, yp-20);
                glVertex2i(xp,    yp-20);
              glEnd();    
              // Border.
              glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
              glColor3f(1.0, 1.0, 1.0);
              glBegin(GL_QUADS);
                glVertex2i(xp,    yp);
                glVertex2i(xp+20, yp);
                glVertex2i(xp+20, yp-20);
                glVertex2i(xp,    yp-20);
              glEnd();
              // Number.
              sprintf(number, "%d", PuzzleBoard[j][i]);
              TEXT_8_X_13_WHITE(xp+1, yp-20+5, number);
           }
       }
   }
   glutSwapBuffers();
}


/***********************************************************/
/* CALLBACK FUNTIONS.                                      */
/***********************************************************/
/****************************************/
/* Mouse control.                       */
/****************************************/
void myMouse(int button, int state, int x, int y) {
int filePicked;
   y = (winHeight - y);
   if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
      if (NEWG_BUTTON.SELECTED(x, y)) { 
         NEWG_BUTTON.CLICK();
         glFinish();
         if (loadFlag == 1) RECOVER_RECORD();
         winFlag  = 0;  
         playFlag = 0;
         loadFlag = 0;
         scanFlag = 1;
         puzzles  = SCAN_FOR_IMAGES("ppm");
         SCAN_WINDOW();
         SCAN_FILE_DISPLAY(scanStart, puzzles);
      }
      if (LVLP_BUTTON.SELECTED(x, y)) { 
         if (playFlag == 0 && loadFlag == 1) {
            LVLP_BUTTON.CLICK();
            difficulty++;
            if (difficulty > 10) difficulty = 10;
            TEXT_DIFFICULTY();
         }
      }
      if (LVLN_BUTTON.SELECTED(x, y)) { 
         if (playFlag == 0 && loadFlag == 1) {
            LVLN_BUTTON.CLICK();
            difficulty--;
            if (difficulty < 0) difficulty = 0;
            TEXT_DIFFICULTY();
         }
      }
      if (PLAY_BUTTON.SELECTED(x, y)) { 
         if (loadFlag == 1) {
            PLAY_BUTTON.CLICK();
            playFlag = 1;
            NEW_CUTUP_IMAGE();
            NEW_RANDOMIZE_IMAGE();
         }
      }
      if (LOAD_BUTTON.SELECTED(x, y)) { 
         LOAD_BUTTON.CLICK();
         HELP_WINDOW.CLEAR(); 
         PUZZLE_LOAD();
         playFlag = 1;  
         TEXT_MOVE_COUNT();
         TEXT_DIFFICULTY();
      }
      if (SAVE_BUTTON.SELECTED(x, y)) { 
         if (playFlag == 1) {
            SAVE_BUTTON.CLICK();
            PUZZLE_SAVE(PuzzleName);
         }
      }
      if (EXIT_BUTTON.SELECTED(x, y)) { 
         EXIT_BUTTON.CLICK();
         glutDestroyWindow(window);
         exit(0);                   
      }
      if (LABL_BUTTON.SELECTED(x, y)) { 
         if (playFlag == 1) {
            LABL_BUTTON.CLICK();
            HELP_NUMBER();
         }
      }
      if (VIEW_BUTTON.SELECTED(x, y)) { 
         if (playFlag == 1) {
            VIEW_BUTTON.CLICK();
            HELP_VIEW();
         }
      }

      // Puzzle piece selection.
      if (x >= 116 && x <= 635 && y >= 10 && y <= 434) {
         if (playFlag == 1 && winFlag == 0) {
            PLAY(x, y);
         }
      }

      // Puzzle name scroll.
      if (UP___BUTTON.SELECTED(x, y) && scanFlag == 1) {
         UP___BUTTON.CLICK();
         if (scanStart <= (puzzles-18)) {
            scanStart++;
            SCAN_FILE_DISPLAY(scanStart, puzzles);
         }
      }
      if (DOWN_BUTTON.SELECTED(x, y)) {
         DOWN_BUTTON.CLICK();
         if (scanStart > 0) {
            scanStart--;
            SCAN_FILE_DISPLAY(scanStart, puzzles);
         }
      }
      if (CNCL_BUTTON.SELECTED(x, y)) { 
         CNCL_BUTTON.CLICK();
         RECOVER_DRAW();
         playFlag = 1;
         loadFlag = 1;
         scanFlag = 0;
      }
      // Puzzle name selection.
      if (x >= 235 && x <= 410 && scanFlag == 1) {
         filePicked = -1;
         if (y >= 380 && y <= 400) filePicked = scanStart;
         if (y >= 360 && y <= 380) filePicked = scanStart+1;
         if (y >= 340 && y <= 360) filePicked = scanStart+2;
         if (y >= 320 && y <= 340) filePicked = scanStart+3;
         if (y >= 300 && y <= 320) filePicked = scanStart+4;
         if (y >= 280 && y <= 300) filePicked = scanStart+5;
         if (y >= 260 && y <= 280) filePicked = scanStart+6;
         if (y >= 240 && y <= 260) filePicked = scanStart+7;
         if (y >= 220 && y <= 240) filePicked = scanStart+8;
         if (y >= 200 && y <= 220) filePicked = scanStart+9;
         if (y >= 180 && y <= 200) filePicked = scanStart+10;
         if (y >= 160 && y <= 180) filePicked = scanStart+11;
         if (y >= 140 && y <= 160) filePicked = scanStart+12;
         if (y >= 120 && y <= 140) filePicked = scanStart+13;
         if (y >= 100 && y <= 120) filePicked = scanStart+14;
         if (y >=  80 && y <= 100) filePicked = scanStart+15;
         if (y >=  60 && y <=  80) filePicked = scanStart+16;
         if (y >=  40 && y <=  60) filePicked = scanStart+17;
         if (filePicked != -1) {
            if (filePicked > puzzles) return;
            PLAY_WINDOW.CLEAR();
            HELP_WINDOW.CLEAR(); 
            strcpy(PuzzleName, PuzzleNames[filePicked]);
            if (NEW_LOAD(PuzzleName) == 0) {
               loadFlag     = 1;
               movesToSolve = 0;
               movesMade    = 0;
               TEXT_MOVE_COUNT();
               difficulty   = 1;
               TEXT_DIFFICULTY();
               NEW_DRAW();
               NEW_HELPVIEW_RECORD();
            }
            scanFlag = 0;
         }
      }
   }
}

/****************************************/
/* Display.                             */
/****************************************/
void myDisplay(void) {

   if (drawFlag == 0) {
      glutSwapBuffers();
   } else {
      MAIN_WINDOW();
      drawFlag = 0;
   }
}

/****************************************/
/* Reshape.                             */
/****************************************/
void myResize(int h, int w){
   if (drawFlag == 0) {
      glutDestroyWindow(window);
      exit(0);                   
   }
}

/****************************************/
/* Init.                                */
/****************************************/
void myInit(void) {
   glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
   glPointSize(1);
   gluOrtho2D(0.0, 640.0, 0.0, 480.0);         
   glShadeModel(GL_SMOOTH);                      
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();                             
   glMatrixMode(GL_MODELVIEW);
   glClear(GL_COLOR_BUFFER_BIT);
}

/***********************************************************/
/* MAIN.                                                   */
/***********************************************************/
int main(int argc, char **argv) {
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);  
   glutInitWindowSize(640, 480);
   glutInitWindowPosition(100, 100);
   window = glutCreateWindow("Puzz-R.A.M. by Ron Marsh");
   glutDisplayFunc(myDisplay);
   glutIdleFunc(PLAY_WINNER);
 //  glutReshapeFunc(myResize);
   glutReshapeFunc(NULL);
   glutMouseFunc(myMouse);
   myInit();
   glutMainLoop();
   return 0;
}


